#!/usr/bin/perl -w

# Copyright 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.010;
use strict;
use warnings;
use FindBin;
use List::Util 'sum';
use Math::PlanePath::AlternateTerdragon;

BEGIN { unshift @INC, $FindBin::Bin }
use MyLineSegments;
use MyPrintwrap ':all';
use MyTerdragon 'dxdy_to_dir3', 'xy_rotate_minus120',
  '@dir6_to_dx', '@dir6_to_dy';

# uncomment this to run the ### lines
# use Smart::Comments;

my $k_max = $ARGV[0] || 8;
print "% generated by tools/$FindBin::Script $k_max\n";

my $path = Math::PlanePath::AlternateTerdragon->new;
my %cache;
sub path_xyxy_to_n_either {
  my ($x1,$y1, $x2,$y2) = @_;
  my $key = "$x1,$y1,$x2,$y2";
  if (! exists $cache{$key}) {
    $cache{$key} = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
  }
  return $cache{$key};
}

{
  my @AltRTS_total;
  my @AltRTS_by_d = ([],[],[]);

  foreach my $k (0 .. $k_max) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my @counts = (0,0,0);
    foreach my $n (0 .. $n_hi-1) {
      ### $n
      my ($x,$y) = $path->n_to_xy($n);
      my ($dx,$dy) = $path->n_to_dxdy($n);
      my ($x2,$y2) = ($x+$dx, $y+$dy);

      my $dir3 = dxdy_to_dir3($dx,$dy);

      my $a_dir3 = ($dir3 - 1) % 3;
      my $b_dir3 = ($a_dir3 - 1) % 3;
      ### dirs: "$dir3, $a_dir3, $b_dir3"

      my ($a_dx,$a_dy) = xy_rotate_minus120($dx,$dy);
      my ($x3,$y3) = ($x2+$a_dx, $y2+$a_dy);

      my $a_n = path_xyxy_to_n_either($x2,$y2, $x3,$y3);
      if (defined $a_n && $a_n >= $n_hi) { undef $a_n; }
      my $b_n = path_xyxy_to_n_either($x3,$y3, $x,$y);
      if (defined $b_n && $b_n >= $n_hi) { undef $b_n; }

      if (defined $a_n && defined $b_n) {
        # 3 side
      } elsif (! defined $a_n && ! defined $b_n) {
        # 1 side
        $counts[$dir3]++;
      } elsif (defined $a_n && ! defined $b_n) {
        # 2 side, b missing
        if ($n < $a_n) {   # count once only for the 2 sides
          ### count2: $b_dir3
          $counts[$b_dir3]++;
        }
      } elsif (! defined $a_n && defined $b_n) {
        # 2 side, a missing
        if ($n < $b_n) {   # count once only for the 2 sides
          ### count2: $a_dir3
          $counts[$a_dir3]++;
        }
      } else {
        die;
      }
    }

    push @AltRTS_total, sum(@counts);
    foreach my $dir3 (0..2) {
      push @{$AltRTS_by_d[$dir3]}, $counts[$dir3];
    }
  }

  print "% GP-DEFINE  gen_AltRTS_total_samples = [",
    join(',',@AltRTS_total),"];\n";

  print "% GP-DEFINE  gen_AltRTS_by_d_samples = {[\n";
  foreach my $dir3 (0..2) {
    print "% GP-DEFINE    ",
      join(',',@{$AltRTS_by_d[$dir3]}),
      ($dir3==2 ? " ]};" : ";"), "\n";
  }
}

{
  my @AltSD_total;
  my @AltSD_by_d = ([],[],[]);

  foreach my $k (0 .. $k_max) {
    ### $k
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my @counts = (0,0,0);
  N: foreach my $n (0 .. $n_hi-1) {
      ### $n
      my ($x,$y) = $path->n_to_xy($n);
      my ($dx,$dy) = $path->n_to_dxdy($n);

      my $num_got = 0;
      my @got = (0) x 6;
      my $dir3 = dxdy_to_dir3($dx,$dy);
      foreach my $dir6 (0 .. 5) {
        my $n2 = path_xyxy_to_n_either
          ($x,$y, $x + $dir6_to_dx[$dir6], $y + $dir6_to_dy[$dir6]);
        if (defined $n2 && $n2 < $n_hi) {
          if ($dir6%2==0 && $n2 < $n) { next N; }
          $got[$dir6] = 1;
          $num_got++;
        }
      }
      ### @got

      if ($num_got == 2) {
        # single-visited, excluding start,end
        my $seen = 0;
        foreach my $dir3 (0 .. 2) {
          if ($got[2*$dir3] && $got[(2*$dir3+1)%6]) {
            ### out following, is inc: $dir3
            #      2*dir3 + 1
            #    .     /
            #     .   /
            #      . v
            #       * ---->  2*dir3
            #        .
            #         .
            $seen++;
            $counts[($dir3+1)%3]++;
          }
          if ($got[2*$dir3] && $got[(2*$dir3-1)%6]) {
            ### out following, is dec: $dir3
            #          .
            #         .
            #        .
            #       * ---->  2*dir3
            #      . ^
            #     .   \
            #    .     \
            #        2*dir3 - 1
            $seen++;
            $counts[($dir3-1)%3]++;
          }
        }
        $seen==1 or die "oops";

      } elsif ($num_got == 4) {
        # double-visited
        my $seen = 0;
        foreach my $dir3 (0 .. 2) {
          if ($got[2*$dir3] && $got[(2*$dir3 + 3)%6]) {
            $seen++;
            $counts[$dir3]++;
          }
        }
        $seen==1 or die "oops";
      }
    }

    push @AltSD_total, sum(@counts);
    foreach my $dir3 (0..2) {
      push @{$AltSD_by_d[$dir3]}, $counts[$dir3];
    }
  }

  print "% GP-DEFINE  gen_AltSD_total_samples = [",
    join(',',@AltSD_total),"];\n";

  print "% GP-DEFINE  gen_AltSD_by_d_samples = {[\n";
  foreach my $dir3 (0..2) {
    print "% GP-DEFINE    ",
      join(',',@{$AltSD_by_d[$dir3]}),
      ($dir3==2 ? " ]};" : ";"), "\n";
  }
}

{
  my @AltLines;
  my @AltLines_by_d = ([],[],[]);
  my @AltLines_lengths_by_d = ([],[],[]);

  foreach my $k (0 .. $k_max) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my @counts = (0,0,0);
    my @by_length = ([],[],[]);
    foreach my $n (0 .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      my ($dx,$dy) = $path->n_to_dxdy($n);

      my $x_before = $x - $dx;
      my $y_before = $y - $dy;
      my $n_before = path_xyxy_to_n_either($x,$y, $x_before,$y_before);
      ### before: "$x_before,$y_before n=".($n_before // 'undef')
      next if defined $n_before && $n_before <= $n_hi;

      # $x,$y has no segment preceding it in direction $dx,$dy, so is the
      # start of a line in that direction
      my $dir3 = dxdy_to_dir3($dx,$dy);
      $counts[$dir3]++;

      my $len = 0;
      for (;;) {
        my $n_after = path_xyxy_to_n_either($x,$y, $x+$dx,$y+$dy);
        last unless defined $n_after && $n_after <= $n_hi;
        $x += $dx;
        $y += $dy;
        $len++;
      }
      $by_length[$dir3]->[$len]++;
    }

    push @AltLines, sum(@counts);
    foreach my $dir3 (0..2) {
      push @{$AltLines_by_d[$dir3]}, $counts[$dir3];
      push @{$AltLines_lengths_by_d[$dir3]}, $by_length[$dir3];
    }
  }

  print "% GP-DEFINE  gen_AltLines_samples = [",
    join(',',@AltLines),"];\n";

  print "% GP-DEFINE  gen_AltLines_by_d_samples = {[\n";
  foreach my $dir3 (0..2) {
    print "% GP-DEFINE    ",
      join(',',@{$AltLines_by_d[$dir3]}),
      ($dir3==2 ? " ]};" : ";"), "\n";
  }

  print "% GP-DEFINE  \\\\ gen_AltLines_lengths_by_d_samples[d+1,k+1][len]\n";
  print "% GP-DEFINE  \\\\ is count how many lines of length \"len\"\n";
  print "% GP-DEFINE  gen_AltLines_lengths_by_d_samples = {[\n";
  foreach my $dir3 (0..2) {
    print "% GP-DEFINE    ",
      join(', ', map { my $aref = $_ // [];
                       foreach my $value (@$aref) { $value //= 0 }
                       "[".join(',',@$aref)."]"}
           @{$AltLines_lengths_by_d[$dir3]}),
             ($dir3==2 ? " ]};" : ";"), "\n";
  }
}
print "% end generated\n";
exit 0;
