#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use FindBin;
use Math::Geometry::Planar;
use Math::PlanePath::TerdragonCurve 117; # v.117 for level_to_n_range()

BEGIN { unshift @INC, $FindBin::Bin }
use MyTerdragon 'transform_to_horizontal',
  'xy_rotate_plus60','xy_rotate_minus60', 'xy_rotate_180',
  'triangle_normalize', 'hull_points_sort_by_angle',
  'Lpred_k','Rpred';
use MyLineSegments;
use MyPrintwrap ':all';


my $level = 6;

print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::TerdragonCurve->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);

sub transform {
  my ($x,$y) = @_;
  return transform_to_horizontal($x,$y, $level);
}
sub transform_to_print {
  my ($x,$y) = @_;
  return ("$x/2", "{$y*sqrt(3)/2}");
}

print "  \\begin{scope}[xscale=1/2,yscale=sqrt(3)/2]\n";

{
  my ($x,$y) = transform(0,0);
  print "  \\coordinate (start) at ($x,$y);\n";
}
{
  my ($x,$y) = transform($x_hi,$y_hi);
  print "  \\coordinate (end) at ($x,$y);\n";
}

my $points = [];
{
  my $segments = MyLineSegments->new;
  foreach my $n ($n_lo .. $n_hi-1) {
    $segments->add($path->n_to_xy($n),
                   $path->n_to_xy($n+1));
  }
  $Printwrap = 2;
  printwrap("\\draw[my grey] ");
  $Printwrap = 4;
  $segments->printwrap(transform => \&transform);
  printwrap(";\n");
}

print "  \\end{scope}\n";

my %triangles;
sub triangle {
  my ($n, $rot) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my $dx = $x2 - $x1;
  my $dy = $y2 - $y1;
  ($dx,$dy) = ($rot >= 0 ? xy_rotate_plus60($dx,$dy)
               : xy_rotate_minus60($dx,$dy));
  my ($x3,$y3) = ($x1+$dx, $y1+$dy);

  push @$points, [$x1,$y1], [$x2,$y2], [$x3,$y3];

  ($x1,$y1) = transform($x1,$y1);
  ($x2,$y2) = transform($x2,$y2);
  ($x3,$y3) = transform($x3,$y3);
  ($x1,$y1, $x2,$y2, $x3,$y3) = triangle_normalize($x1,$y1, $x2,$y2, $x3,$y3);

  my $above = ($y3 > $y1 ? "above" : $y3 < $y1 ? "below" : die);
  ($x1,$y1) = transform_to_print($x1,$y1);
  push @{$triangles{$above}}, "($x1,$y1)";
}
{
  foreach my $n ($n_lo .. $n_hi-1) {
    my $rot = 0;
    if (Lpred_k($n,$level)) { triangle($n,1);  }
    if (Rpred($n,$level)) {   triangle($n,-1); }
  }

  foreach my $above ('above','below') {
    $Printwrap = 2;
    printwrap("\\fill foreach \\p in {");
    $Printwrap += 2;
    foreach my $xy (@{$triangles{$above}}) {
      printwrap("$xy,");
    }
    printwrap("}");
    printwrap(" { \\p pic[transform shape] {my triangle $above pic} };\n");
  }
}

my $planar = Math::Geometry::Planar->new;
$planar->points($points);
if (@$points >= 5) {
  $planar = $planar->convexhull2;
  $points = $planar->points;
}
$points = hull_points_sort_by_angle($points, xy_rotate_180($x_hi,$y_hi));
unshift @$points, pop @$points;


my @point_to_PT;
if ($level == 4) {
  @point_to_PT = ([1,2,3],[4],[5],[6,7]);
} else {
  @point_to_PT = ([1],[2],[3],[4],[5],[6],[7]);
}
scalar(@$points) == 2*scalar(@point_to_PT)
  || die "points ",scalar(@$points)," PT array ",scalar(@point_to_PT);

my $num = 1;
my $dash = '';
foreach my $i (0 .. $#$points) {
  my $p = $points->[$i];
  my ($x,$y) = @$p;
  my $dash = ($i<=$#point_to_PT ? '' : 'dash');
  my $aref = $point_to_PT[$i<=$#point_to_PT ? $i : $i-scalar(@point_to_PT)];
  my $num = $aref->[0];
  ($x,$y) = transform($x,$y);
  ($x,$y) = transform_to_print($x,$y);
  print "  \\node (PT$num$dash) at ($x,$y) [my dot];\n";
  $num++;
  if ($num == 8) {
    $num = 1;
    $dash = 'dash';
  }
}

$num = 1;
$dash = '';
foreach my $i (0 .. $#$points) {
  my $p = $points->[$i];
  my ($x,$y) = @$p;
  my $dash = ($i<=$#point_to_PT ? '' : 'dash');
  my $aref = $point_to_PT[$i<=$#point_to_PT ? $i : $i-scalar(@point_to_PT)];
  foreach my $num (@$aref) {
    print "% GP-Test-Last  PT$num$dash($level) == $x/2 + $y*sqrt3i/2\n";
  }
}

$Printwrap = 2;
printwrap("\\draw[dashed] ");
$Printwrap += 2;
my $join = '';
foreach my $dash ('', 'dash') {
  foreach my $i (0 .. $#point_to_PT) {
    my $num = $point_to_PT[$i]->[0];
    printwrap("$join(PT$num$dash)");
    $join = '--';
  }
}
printwrap("$join(PT1);\n");

print "  % end generated\n";
exit 0;
